/*===========================================================================
 * Licensed Materials - Property of IBM
 * "Restricted Materials of IBM"
 * 
 * IBM SDK, Java(tm) Technology Edition, v8
 * (C) Copyright IBM Corp. 1995, 2022. All Rights Reserved
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *===========================================================================
 */
/*
 * Copyright (c) 1995, 2011, Oracle and/or its affiliates. All rights reserved.
 * ORACLE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 */

/**
 * =======================================================================
 * Module Information:
 *
 * DESCRIPTION: This class implements RDMA socket by extending from
 * the PlainSockImpl.
 * 
 * This package private class provides RDMA_CAPABILITY.
 * =======================================================================
 */

package java.net;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.FileDescriptor;
import java.security.AccessController;
import sun.security.action.GetPropertyAction;
import sun.security.action.GetBooleanAction;

class RDMASocketImpl extends PlainSocketImpl {
	// associated input/output streams 
	private RDMASocketInputStream socketInputStream = null;
	private RDMASocketOutputStream socketOutputStream = null;
	// store the pointer of the native data structure related to this socket
	private long socketDataPtr = 0;
	Object socketDataPtrLock = new Object();
	private final SocketImpl oldImpl;
	private boolean created = false;
	private final boolean zCopy;
	private final int zCopyThreshold;

	// tcp fallback support
	private static final boolean tcpFallback;
	private static final int DFLT_TCP_FALLBACK_WAIT_TIME = 100; // milliseconds
	private static final int MIN_TCP_FALLBACK_WAIT_TIME = 20; // milliseconds
	private static final int MAX_TCP_FALLBACK_WAIT_TIME = 10000; // milliseconds
	private static final int tcpFallbackWaitTime;
	private boolean tcpFallbackInPlace = false;
	private static final String connTimedOutStr = "Connection timed out";
	private static final String acceptTimedOutStr = "Accept timed out";
	private static final boolean nativeForkCompatibility;
	private static boolean safeSocketOps = false; // thread-safe read/write/close operations on RDMA socket

	// native methods
	static native void initProtoRDMA();
	native long socketCreateRDMA(FileDescriptor fdObj) throws IOException;
	native void socketConnectRDMA(InetAddress address, int port, long socketDataPtr,
			int timeOut) throws IOException;
	native void socketBindRDMA(InetAddress address, int port, long socketDataPtr)
			throws IOException;
	native void socketListenRDMA(int count, long socketDataPtr) throws IOException;
	native long socketAcceptRDMA(SocketImpl acceptedSocketImpl,
			long serverSocketDataPtr, int timeOut) throws IOException;
	native void socketCloseRDMA(boolean useDeferredClose, long socketDataPtr)
			throws IOException;
	native void socketShutdownRDMA(int howto, long socketDataPtr) throws IOException;
	native void socketSendUrgentDataRDMA(int data, long socketDataPtr) throws IOException;
	native void socketSetOptionRDMA(int cmd, boolean on, Object value,
			long socketDataPtr) throws SocketException;
	native int socketGetOptionRDMA(int opt, Object iaContainerObj,
			long socketDataPtr) throws SocketException;
	native int socketAvailableRDMA(long socketDataPtr) throws IOException;

	// do native initialization only once
	static {
		boolean fallback = false;
		String tcpFallbackPropStr = null;
		String tcpFallbackWaitTimePropStr = null;
		int fallbackWaitTime = DFLT_TCP_FALLBACK_WAIT_TIME;
		boolean compatibility = false;
		String nativeForkCompatibilityPropStr = null;

		initProtoRDMA();

		tcpFallbackPropStr = AccessController.doPrivileged(
				new GetPropertyAction("com.ibm.net.rdma.tcpFallback"));
		if (tcpFallbackPropStr != null) {
			fallback = Boolean.parseBoolean(tcpFallbackPropStr);
		}
		tcpFallback = fallback;

		tcpFallbackWaitTimePropStr = AccessController.doPrivileged(
				new GetPropertyAction("com.ibm.net.rdma.tcpFallbackWaitTime"));
		if (tcpFallbackWaitTimePropStr != null) {
			try {
				fallbackWaitTime = Integer.parseInt(tcpFallbackWaitTimePropStr);
				if (fallbackWaitTime < MIN_TCP_FALLBACK_WAIT_TIME) {
					System.out.printf(
							"JSORW:NET: Invalid tcp fallback wait time \'%d\' - defaulting to %d milliseconds%n",
							fallbackWaitTime, MIN_TCP_FALLBACK_WAIT_TIME);
					fallbackWaitTime = MIN_TCP_FALLBACK_WAIT_TIME;
				} else if (fallbackWaitTime > MAX_TCP_FALLBACK_WAIT_TIME) {
					System.out.printf(
							"JSORW:NET: Invalid tcp fallback wait time \'%d\' - defaulting to %d milliseconds%n",
							fallbackWaitTime, MAX_TCP_FALLBACK_WAIT_TIME);
					fallbackWaitTime = MAX_TCP_FALLBACK_WAIT_TIME;
				}
			} catch (NumberFormatException e) {
				System.out.println(
						"JSORW:NET: Invalid \"com.ibm.net.rdma.tcpFallbackWaitTime\" property value ("
						+e.getMessage()+"): defaulting to "+DFLT_TCP_FALLBACK_WAIT_TIME+" milliseconds");
				fallbackWaitTime = DFLT_TCP_FALLBACK_WAIT_TIME;
			}
		}
		tcpFallbackWaitTime = fallbackWaitTime;

		nativeForkCompatibilityPropStr = AccessController.doPrivileged(
				new GetPropertyAction("com.ibm.net.rdma.nativeForkCompatibility"));
		if (nativeForkCompatibilityPropStr != null) {
			compatibility = Boolean.parseBoolean(nativeForkCompatibilityPropStr);
		}
		nativeForkCompatibility = compatibility;
		safeSocketOps = AccessController.doPrivileged(
		        new GetBooleanAction("com.ibm.net.rdma.threadSafeSocketOperations")).booleanValue();
	}

	// helper method to create RDMA socket and set data pointer field
	private void socketCreatePrivate(FileDescriptor fdObj) throws IOException {
		socketDataPtr = socketCreateRDMA(fdObj);
		created = true;
		if (NetworkSelector.isDebugOn()) {
			System.out.printf(
					"JSORI:NET:%X: RDMA socket created%n",
					this.hashCode());
		}
	}

	/**
	 * Constructs a RDMASocketImpl instance
	 */
	RDMASocketImpl(SocketImpl oldImpl, boolean zCopy, int zCopyThreshold) {
		this.oldImpl = oldImpl;
		this.zCopy = zCopy;
		this.zCopyThreshold = zCopyThreshold;
	}

	/**
	 * Creates an RDMA stream socket
	 */
	@Override
	protected void create(boolean stream) throws IOException {
		if (!stream) {
			throw new UnsupportedOperationException(
					"Must be a stream socket for RDMA!");
		}
		super.create(stream);
	}


	/**
	 * Return true if this RDMA socket is already closed or close is pending
	 */
	@Override
	public boolean isClosedOrPending() {
		if (tcpFallbackInPlace) {
			return super.isClosedOrPending();
		} else {
			if (socketDataPtr == 0) {
				return true;
			} else {
				return false;
			}
		}
	}

	/**
	 * Get input stream for this RDMA socket.
	 */
	@Override
	protected synchronized InputStream getInputStream() throws IOException {
		if (tcpFallbackInPlace) {
			return super.getInputStream();
		} else {
			if (isClosedOrPending()) {
				throw new IOException("RDMA Socket Closed");
			}
			if (inputShut()) {
				throw new IOException("RDMA Socket input is shutdown");
			}
			if (socketInputStream == null) {
				socketInputStream = new RDMASocketInputStream(this);
				if (NetworkSelector.isDebugOn()) {
					System.out.printf(
							"JSORI:NET:%X: RDMA socket input stream created%n",
							this.hashCode());
				}
			}
			return socketInputStream;
		}
	}

	/**
	 * Get output stream for the RDMA socket.
	 */
	@Override
	protected synchronized OutputStream getOutputStream() throws IOException {
		if (tcpFallbackInPlace) {
			return super.getOutputStream();
		} else {
			if (isClosedOrPending()) {
				throw new IOException("RDMA Socket Closed");
			}
			if (outputShut()) {
				throw new IOException("RDMA Socket output is shutdown");
			}
			if (socketOutputStream == null) {
				socketOutputStream = new RDMASocketOutputStream(this);
				if (NetworkSelector.isDebugOn()) {
					System.out.printf(
							"JSORI:NET:%X: RDMA socket output stream created%n",
							this.hashCode());
				}
			}
			return socketOutputStream;
		}
	}

	/**
	 * z/OS specific function hence null implementation to satisfy base class
	 */
	void processNRException(InetAddress address, int port, int backlogQ) {
		// do nothing
	}

	/**
	 * Create an RDMA stream socket
	 */
	@Override
	void socketCreate(boolean isServer) throws IOException {
		super.socketCreate(isServer);
		socketCreatePrivate(null);
	}

	/**
	 * Connect this RDMA socket to the specified port on the specified host
	 */
	@Override
	void socketConnect(InetAddress address, int port, int timeout) throws IOException {
		/* The socket connect can happen lately after creating a plain unconnected socket.
		 * In such a scenario, we need to ensure that RDMA portion of the socket is also
		 * created and linked to the plain socket before attempting the connect operation.
		 */
		if (!created && (oldImpl != null)) {
			socketCreatePrivate(oldImpl.getFileDescriptor());
		}

		/* hack to make jsor to rsockets interoperability work
		 * rsockets lazily converts a normal socket into rsocket
		 * during connection acceptance - so the client should start
		 * connection request on a normal socket before attempting
		 * on rsocket
		 */
		if (nativeForkCompatibility && !tcpFallback) {
			try {
				super.socketConnect(address, port, timeout);
			} catch (Exception e) {
				// Ignore the failure
			}
		}

		try {
			int waitTime = 0;
			if (timeout < 0) {
				System.out.printf(
						"JSORW:NET:%X: Invalid timeout parameter \'%d\' - defaulting to zero (infinite timeout)%n",
						this.hashCode(), timeout);
				timeout = 0;
				waitTime = timeout;
			} else if (timeout > 0) {
				waitTime = timeout;
			} else {
				waitTime = 0;
			}

			socketConnectRDMA(address, port, socketDataPtr, waitTime);
			if (NetworkSelector.isDebugOn()) {
				System.out.printf(
						"JSORI:NET:%X: RDMA connection established to %s:%d%n",
						this.hashCode(), address, port);
			}
		} catch (Exception e) {
			socketCloseRDMA(false, socketDataPtr);
			socketDataPtr = 0;
			if (e instanceof SocketTimeoutException) {
				throw new SocketTimeoutException(e.getMessage());
			} else {
				throw new IOException(e.getMessage());
			}
		}
	}

	/**
	 * Bind this RDMA socket to the specified IP address and port number
	 */
	@Override
	void socketBind(InetAddress address, int port) throws IOException {
		/* The socket bind can happen lately after a plain server socket creation.
		 * If that is the case, then we need to ensure that RDMA portion of the server
		 * socket is also created and linked to the plain server socket before attempting
		 * the bind operation.
		 */
		if (!created && (oldImpl != null)) {
			socketCreatePrivate(oldImpl.getFileDescriptor());
		}

		super.socketBind(address, port);
		socketBindRDMA(address, port, socketDataPtr);
		if (NetworkSelector.isDebugOn()) {
			System.out.printf(
					"JSORI:NET:%X: RDMA socket bound to %s:%d%n",
					this.hashCode(), address, port);
		}
	}

	/**
	 * Listen for incoming connection indications on this RDMA socket
	 * with the maximum queue length of 'count'
	 */
	@Override
	void socketListen(int count) throws IOException {
		if (tcpFallback) {
			super.socketListen(count);
		}
		socketListenRDMA(count, socketDataPtr);
		if (NetworkSelector.isDebugOn()) {
			System.out.printf(
					"JSORI:NET:%X: RDMA socket now in listening mode%n",
					this.hashCode());
		}
	}

	/**
	 * Accept the incoming connections on the RDMA server socket
	 */
	@Override
	void socketAccept(SocketImpl s) throws IOException {
		long clientSocketDataPtr = 0;

		if (!(s instanceof RDMASocketImpl)) {
			return;
		}

		try {
			int waitTime = 0;
			int acceptTimeout = 0;

			acceptTimeout = ((PlainSocketImpl)this).getTimeout();
			if (acceptTimeout == 0 && (oldImpl != null)) {
				acceptTimeout = ((PlainSocketImpl)oldImpl).getTimeout();
			}

			if (tcpFallback) {
				waitTime = tcpFallbackWaitTime;
			} else if (acceptTimeout > 0) {
				waitTime = acceptTimeout;
			} else {
				waitTime = 0;
			}

			clientSocketDataPtr = socketAcceptRDMA(s, socketDataPtr, waitTime);
			if (NetworkSelector.isDebugOn()) {
				System.out.printf(
						"JSORI:NET:%X: Accepted RDMA connection request on %X%n",
						this.hashCode(), s.hashCode());
			}
		} catch (Exception e) {
			if (e instanceof SocketTimeoutException) {
				if (tcpFallback) {
					super.socketAccept((PlainSocketImpl)s);
					((RDMASocketImpl)s).setTcpFallbackInPlace(true);
					clientSocketDataPtr = 0;
				} else {
					throw new SocketTimeoutException(e.getMessage());
				}
			} else {
				throw new IOException(e.getMessage());
			}
		}

		((RDMASocketImpl)s).setSocketDataPtr(clientSocketDataPtr);
	}

	/**
	 * Close this RDMA socket
	 */
	@Override
	void socketClose0(boolean useDeferredClose) throws IOException {
            if (safeSocketOps) {
                synchronized(socketDataPtrLock) {
                    socketCloseImpl(useDeferredClose);
                }
            } else {
                socketCloseImpl(useDeferredClose);
            }
	}

	private void socketCloseImpl(boolean useDeferredClose) throws IOException {
                // close the RDMA socket followed by the plain socket
                if (socketDataPtr != 0) {
                        socketCloseRDMA(useDeferredClose, socketDataPtr);
                        if (!useDeferredClose) {
                                socketDataPtr = 0;
                                if (NetworkSelector.isDebugOn()) {
                                        System.out.printf(
                                                        "JSORI:NET:%X: RDMA socket closed%n",
                                                        this.hashCode());
                                }
                        }
                }
                super.socketClose0(useDeferredClose);
        }

	/**
	 * Shutdown this RDMA connection
	 */
	@Override
	void socketShutdown(int howto) throws IOException {
            if(safeSocketOps) {
                synchronized (socketDataPtrLock) {
	            socketShutdownImpl(howto);
		}
            } else {
                socketShutdownImpl(howto);
            }
	}

	private void socketShutdownImpl(int howto) throws IOException {
                // shutdown the RDMA/plain socket
                if (tcpFallbackInPlace) {
                        super.socketShutdown(howto);
                } else {
                        socketShutdownRDMA(howto, socketDataPtr);
                        if (NetworkSelector.isDebugOn()) {
                                System.out.printf(
                                                "JSORI:NET:%X: RDMA socket closed for %s%n",
                                                this.hashCode(),
                                                ((howto == 0) ? "read" : ((howto == 1) ? "write" : "read/write")));
                        }
                }
        }

	/**
	 * Send urgent data on this RDMA socket connection
	 */
	@Override
	void socketSendUrgentData(int data) throws IOException {
		if (tcpFallbackInPlace) {
			super.socketSendUrgentData(data);
		} else {
			socketSendUrgentDataRDMA(data, socketDataPtr);
		}
	}

	/**
	 * Returns the local port to which this socket is bound.
	 */
	@Override
	protected int getLocalPort() {
		if (tcpFallbackInPlace) {
			return super.getLocalPort();
		} else {
			if (NetworkSelector.isDebugOn()) {
				System.out.printf(
						"JSORI:NET:%X: Retrieved RDMA socket local port %d%n",
						this.hashCode(), localport);
			}
			return localport;
		}
	}

	/**
	 * Returns the remote port to which this socket is connected.
	 */
	@Override
	protected int getPort() {
		if (tcpFallbackInPlace) {
			return super.getPort();
		} else {
			if (NetworkSelector.isDebugOn()) {
				System.out.printf(
						"JSORI:NET:%X: Retrieved RDMA socket remote port %d%n",
						this.hashCode(), port);
			}
			return port;
		}
	}

	/**
	 * Returns the address to which the socket is connected.
	 */
	@Override
	protected InetAddress getInetAddress() {
		if (tcpFallbackInPlace) {
			return super.getInetAddress();
		} else {
			if (NetworkSelector.isDebugOn()) {
				System.out.printf(
						"JSORI:NET:%X: Retrieved RDMA socket remote addr %s%n",
						this.hashCode(), address);
			}
			return address;
		}

	}

	/**
	 * socketSetOption()
	 */
	@Override
	protected void socketSetOption(int cmd, boolean on,
			Object value) throws SocketException {
		if (tcpFallbackInPlace) {
			super.socketSetOption(cmd, on, value);
		} else {
			switch (cmd) {
			case SO_REUSEADDR:
			case SO_RCVBUF:
			case SO_SNDBUF:
			case SO_LINGER:
			case SO_KEEPALIVE:
			case SO_OOBINLINE:
			case TCP_NODELAY:
				socketSetOptionRDMA(cmd, on, value, socketDataPtr);
				if (NetworkSelector.isDebugOn()) {
					System.out.printf(
							"JSORI:NET:%X: RDMA socket option 0x%x set to %s%n",
							this.hashCode(), cmd, on);
				}
				break;
			default:
				// other socket related options are not supported on RDMA socket
				break;
			}
		}
		return;
	}

	/**
	 * socketGetOption()
	 */
	@Override
	int socketGetOption(int opt, Object iaContainerObj)
			throws SocketException {
		int optVal = -1;

		if (tcpFallbackInPlace) {
			optVal = super.socketGetOption(opt, iaContainerObj);
		} else {
			switch (opt) {
			case SO_REUSEADDR:
			case SO_RCVBUF:
			case SO_SNDBUF:
			case SO_LINGER:
			case SO_KEEPALIVE:
			case SO_OOBINLINE:
			case SO_BINDADDR:
			case TCP_NODELAY:
				optVal = socketGetOptionRDMA(opt, iaContainerObj, socketDataPtr);
				if (NetworkSelector.isDebugOn()) {
					System.out.printf(
							"JSORI:NET:%X: RDMA socket option 0x%x value %d%n",
							this.hashCode(), opt, optVal);
				}
				break;
			default:
				// other socket options are not supported on RDMA socket
				optVal = -1;
				break;
			}
		}
		return optVal;
	}

	/**
	 * Method to get available number of bytes to read.
	 * It overrides InputStream.available
	 */
	@Override
	int socketAvailable() throws IOException {
		if (tcpFallbackInPlace) {
			return super.socketAvailable();
		} else {
			return socketAvailableRDMA(socketDataPtr);
		}
	}

	/**
	 * Getter method for the RDMA socket pointer
	 */
	long getSocketDataPtr() {
		return socketDataPtr;
	}

	/**
	 * Setter method for the RDMA socket pointer
	 */
	void setSocketDataPtr(long socketDataPtr) {
		this.socketDataPtr = socketDataPtr;
	}

	/**
	 * Setter method for the tcp fallback status field
	 */
	void setTcpFallbackInPlace(boolean fallback) {
		this.tcpFallbackInPlace = fallback;
		if (NetworkSelector.isDebugOn()) {
			System.out.printf(
					"JSORI:NET:%X: TCP fallback mode in place for RDMA socket%n",
					this.hashCode());
		}
	}

	/**
	 * Getter method for the zero copy enablement field
	 */
	boolean isZeroCopy() {
		return zCopy;
	}

	/**
	 * Getter method for the zero copy threshold attribute
	 */
	int getZeroCopyThreshold() {
		return zCopyThreshold;
	}
}
