/*===========================================================================
 * Licensed Materials - Property of IBM
 * "Restricted Materials of IBM"
 * 
 * IBM SDK, Java(tm) Technology Edition, v8
 * (C) Copyright IBM Corp. 1997, 2022. All Rights Reserved
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *===========================================================================
 */
/*
 * Copyright (c) 1997, 2011, Oracle and/or its affiliates. All rights reserved.
 * ORACLE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 */

/**
 * =======================================================================
 * Module Information:
 *
 * DESCRIPTION: This class implements RDMA socket input stream by
 * extending SocketInputStream.  Note that this class should <b>NOT<b>
 * be made public.
 * 
 * This package private class provides RDMA_CAPABILITY.
 * =======================================================================
 */

package java.net;

import java.io.FileDescriptor;
import java.io.IOException;
import java.security.AccessController;
import sun.net.ConnectionResetException;
import sun.security.action.GetBooleanAction;

class RDMASocketInputStream extends SocketInputStream {
	// native methods
	private native static void init();
	private native int socketRead0(long sockDataPtr, byte b[],
			int off, int len, int timeout) throws IOException;
	private native int socketRead0Direct(long sockDataPtr, byte b[],
			int off, int len, int timeout) throws IOException;
	
	// end of file flag
	private boolean eof;
	// RDMA socket impl associated with this input stream
	private RDMASocketImpl impl = null;
	
	// Synchronize read/write operations against close/shutdown. This could have a performance cost.
	private static boolean safeSocketOps = false;

	// perform class load-time initialization only once
	static {
		safeSocketOps = AccessController.doPrivileged(
                        new GetBooleanAction("com.ibm.net.rdma.threadSafeSocketOperations")).booleanValue();
		init();
	}
	
	/**
	 * Creates a new RDMA socket input stream.
	 */
	RDMASocketInputStream(RDMASocketImpl impl) throws IOException {
		super(impl);
		this.impl = impl;
	}
	
	/**
	 * Reads into a byte array the specified number of bytes at
	 * the specified offset.
	 */
	int read(byte b[], int off, int length, int timeout)
			throws IOException {
		// read byte count
		int n;
		
		// eof reached?
		if (eof) {
			return -1;
		}
		
		// connection reset
		if (impl.isConnectionReset()) {
			throw new SocketException("RDMA Connection reset");
		}
		
		// bounds check
		if (length <= 0 || off < 0 || (off + length) > b.length) {
			if (length == 0) {
				return 0;
			}
			throw new ArrayIndexOutOfBoundsException();
		}
		
		try {
			if (impl.isZeroCopy() && length >=
					(impl.getZeroCopyThreshold() * 1024)) {
				if (safeSocketOps) {
				    synchronized(impl.socketDataPtrLock) {
					if (impl.getSocketDataPtr() == 0)
					    throw new SocketException("RDMA socket closed");
				        n = socketRead0Direct(impl.getSocketDataPtr(), b, off,
					 	        length, timeout);
				    }
				} else {
				   n = socketRead0Direct(impl.getSocketDataPtr(), b, off,
						   length, timeout);
				}
			} else {
				if (safeSocketOps) {
				    synchronized(impl.socketDataPtrLock) {
					if (impl.getSocketDataPtr() == 0)
					    throw new SocketException("RDMA socket closed");
				        n = socketRead0(impl.getSocketDataPtr(), b, off,
						    length, timeout);
                                    }
                                } else {
				    n = socketRead0(impl.getSocketDataPtr(), b, off,
				                length, timeout);
				}
			}
			if (n > 0) {
				return n;
			}
		} catch (ConnectionResetException cre) {
			impl.setConnectionResetPending();
		}
		
		// the socket might have been closed
		if (impl.isClosedOrPending()) {
			throw new SocketException("RDMA Socket closed");
		}
		
		// the connection could have been reset
		if (impl.isConnectionResetPending()) {
			impl.setConnectionReset();
		}
		if (impl.isConnectionReset()) {
			throw new SocketException("RDMA Connection reset");
		}
		
		// set eof flag
		eof = true;
		return -1;
	}
}
